"""Generic converter that converts an Epub to OpenITI mARkdown.

Examples:
    Generic epub conversion:
    >>> from epub_converter_generic import GenericEpubConverter
    >>> gen_converter = GenericEpubConverter(dest_folder="test/converted")
    >>> gen_converter.VERBOSE = False
    >>> folder = r"test"
    >>> fn = r"Houellebecq 2019 - serotonine.epub"
    >>> gen_converter.convert_file(os.path.join(folder, fn))
    >>> gen_converter.convert_files_in_folder(folder, ["epub",])

    Sub-class to create a converter specific to epubs from the Hindawi library:
    >>> HindawiConverter = GenericEpubConverter("test/converted")
    >>> HindawiConverter.VERBOSE = False
    >>> import html2md_hindawi
    >>> HindawiConverter.convert_html2md = html2md_hindawi.markdownify  # (1)
    >>> HindawiConverter.toc_fn = "nav.xhtml"  # (2)
    >>> folder = r"test"
    >>> fn = r"26362727.epub"
    >>> HindawiConverter.convert_file(os.path.join(folder, fn))
    
    # (1) overwrite the convert_html2md function
    # (2) specify the filename of the table of contents in Hindawi epub files

An Epub file is in fact a zipped archive.
The most important element of the archive for conversion purposes
are the folders that contain the html files with the text.
Some Epubs have a table of contents that defines the order
in which these html files should be read.

The GenericEpubConverter is a subclass of the GenericConverter
from the generic_converter module:

GenericConverter
    |_ GenericEpubConverter

GenericEpubConverter's main methods are inherited from the GenericConverter:
* convert_file(source_fp): basic procedure for converting an epub file.
* convert_files_in_folder(source_folder): convert all epub files in the folder
    (calls convert_file)

Methods of both classes:
(methods of GenericConverter are inherited by GenericEpubConverter;
methods of GenericConverter with the same name
in GenericEpubConverter are overwritten by the latter)

| *GenericConverter*          | *GenericEpubConverter*
|-----------------------------|-----------------------
| __init__                    | __init__
| convert_files_in_folder     | (inherited)
| convert file                | (inherited)
| make_dest_fp                | (inherited - generic!)
| get_metadata                | (inherited - generic!)
| get_data                    | get_data
| pre_process                 | (inherited)
| add_page_numbers            | (inherited - generic!)
| add_structural_annotations  | (inherited - generic!)
| remove_notes                | remove_notes
| reflow                      | (inherited)
| add_milestones              | (inherited)
| post_process                | (inherited - generic!)
| compose                     | (inherited)
| save_file                   | (inherited)
|                             | inspect_epub
|                             | sort_html_files_by_toc
|                             | add_unique_tags

To create a converter for a specific type of epubs,
subclass the GenericEpubConverter and overwrite
some of its methods:

GenericConverter
    |_ GenericEpubConverter
            |_ HindawiEpubConverter
            |_ ShamelaEpubConverter
            |_ ...
"""

import codecs
import os
import re
import zipfile
import time

from bs4 import BeautifulSoup

from generic_converter import GenericConverter
import html2md


class GenericEpubConverter(GenericConverter):

    def __init__(self, dest_folder=None):
        """Initialize the class and its super-class."""
        super().__init__()
        self.root_tag_string = "soup.html.body"
        self.unique_tags = dict()
        if dest_folder == None:
            self.dest_folder = "converted"
        else:
            self.dest_folder = dest_folder
        self.VERBOSE = True


##    def make_dest_fp(self, source_fp):
##        """"""
##        source_folder = os.path.split(source_fp)[0]
##        self.dest_folder = os.path.join(source_folder, "converted")
##        self.dest_fp = super().make_dest_fp(source_fp)

    def inspect_epub(self, src_fp):
        """Print the contents of the zip archive."""
        if self.VERBOSE:
            print("-"*60)
            print("Contents of  {}:".format(src_fp))
        if src_fp.endswith(".epub"):
            zp = zipfile.ZipFile(src_fp)
            if self.VERBOSE:
                for info in zp.infolist():
                    print(info.filename)


    def convert_html2md(self, html):
        """Convert the Epub's internal html to mARkdown text.

        *** USING A GENERIC HTML TO MARKDOWN CONVERTER HERE.
        OVERWRITE THIS FUNCTION IN THE SUB-CLASS
        TO CONVERT THE EPUB'S SPECIFIC HTML FLAVOUR TO MARKDOWN ***
        """
        text = html2md.markdownify(html)
        return text


    def add_unique_tags(self, root_tag, src_fp, fn):
        """Adds tags to the unique_tags_dictionary that were not yet in it.

        Args:
            root_tag (BeautifulSoup object): html element containing the text.
            unique_tabs (dict): a dictionary containing all tags used in the
                converted texts.
            src_fp (str): the path to the file currently being converted.
            fn (str): the filename of the string currently being converted.

        Returns:
            None
        """
        all_tags = root_tag.find_all()
    ##    print("number of tags in document {}: {}".format(fn, len(all_tags)))
    ##    all_divs = root_tag.find_all("div")
    ##    print("number of divs in document {}: {}".format(fn, len(all_divs)))

        for tag in all_tags:
            if tag.name not in self.unique_tags:
                try:
                    class_ = repr(tag["class"])
                except:
                    class_ = "no class"
                self.unique_tags[tag.name]={class_: (src_fp, fn)}
            else:
                try:
                    class_ = repr(tag["class"])
                except:
                    class_ = "no class"
                if class_ not in self.unique_tags[tag.name]:
                    self.unique_tags[tag.name][class_] = (src_fp, fn)


    def sort_html_files_by_toc(self, zp, toc_fp, html_files):
        """Gets the table of contents from the Epub file.

        Args:
            zp: zipfile object
            toc_fp (str): filepath to the table of contents of the epub.
            html_files(list): an unordered list of the html files in the epub.

        Returns:
            toc (list): a list of filepaths to the html files
                in the epub file, in the order specified by the
                table of contents
        """
        html_files_dict = {os.path.split(fp)[-1] : fp for fp in html_files}
        toc_data = zp.read(toc_fp)
        toc_data = codecs.decode(toc_data, "utf-8")
        soup = BeautifulSoup(toc_data)
        toc_ol = soup.find("ol")
        toc = []
        for a in toc_ol.find_all("a"):
            fn = os.path.split(a.get("href"))[-1]
            if fn in html_files_dict:
                toc.append(html_files_dict[fn])
        return toc


    def remove_notes(self, text):
        """Remove the footnotes from the text.

        Args:
            text (str)

        Returns:
            text_without_notes (str): text from which the notes were removed.
            notes (str): the footnotes converted to endnotes
                (if there are endnotes: add the endnote_splitter before them)
        """
        text_without_notes = ""
        notes = ""
        spl = re.split("(FOOTNOTE.*\n+)", text)
        spl = [x for x in spl if x != ""]
        i = 0
        #last_text_fragm = 0
        for t in spl:
            if not t.strip().startswith("FOOTNOTE"):
                text_without_notes += t
                #last_text_fragm = i
            else:
                fn_text = t.strip()[8:]
                if spl[i-1].strip().startswith("FOOTNOTE"):
                    notes += "{}\n\n".format(fn_text)
                else:
                    #p = re.findall("PageV\d+P\d+", spl[last_text_fragm])
                    p = re.findall("PageV\d+P\d+", spl[i-1])
                    if p:
                        notes += "{}:\n\n{}\n\n".format(p[-1], fn_text)
                    else:
                        section_rgx = "(### [\|$]+) (.*)"
                        #sections = re.findall(section_rgx, spl[last_text_fragm])
                        sections = re.findall(section_rgx, spl[i-1])
                        if sections:
                            msg = "Notes on section ({})"
                            msg +=  "(no page numbers):\n\n{}\n\n"
                            notes += msg.format(sections[0][1], fn_text)
                        else:
                            notes += "(no page number):\n\n{}\n\n".format(fn_text)
            i += 1
        if notes:
            notes = self.endnote_splitter + notes
        return text_without_notes, notes


    def get_toc(self, src_fp):
        """Get the filename of the table of contents of the epub"""
        try:
            self.toc_fn
        except:
            self.inspect_epub(src_fp)
            toc_input = """\
Write the filename (with extension) of the table of contents
(if there is no table of contents, simply press Enter): """
            if self.VERBOSE:
                resp = input(toc_input)
            else:
                resp = ""
            #print(resp)
            if resp.endswith("ml"):
                self.toc_fn = resp
            else:
                self.toc_fn = "no toc"
        return self.toc_fn


    def get_data(self, src_fp):
        """Converts an Epub file to an OpenITI mARkdown document.

        If a dictionary is passed as the unique_tags parameter,
        the function will keep track of all tags (and their classes)
        used in the converted Epub file.

        Args:
            src_fp (str): the path to the file currently being converted.
            dest_folder (str): the path to the folder where the converted file
                should be stored.
            conv_func (function): the function used to convert html to mARkdown.
            toc_fn (str): filename of the table of contents file within the epub.
            root_tag_string (str): the path to the BeautifulSoup element that
                contains the full text.
                Default: the body tag.
            unique_tabs (dict): a dictionary containing all tags used in the Epub.
                Default: None (tags used in the epubs will not be listed)

        Returns:
            unique_tabs (dict): a dictionary containing all tags used in the Epub.
        """
        if src_fp.endswith(".epub"):
            zp = zipfile.ZipFile(src_fp)
            if self.VERBOSE:
                fn = os.path.basename(src_fp)
##                msg = "Press Enter to convert %s to txt..." % fn
##                input(msg)

            # make a list of the html files in the epub:

            toc_fn = self.get_toc(src_fp)
            html_files = []
            for info in zp.infolist():
                if info.filename.endswith(self.toc_fn):
                    toc_fp = info.filename
                    if self.VERBOSE:
                        print("toc_fp", toc_fp)
                elif info.filename.endswith("html"):
                    html_files.append(info.filename)

            # sort the list of the html files in the correct order:

            try:
                html_files = self.sort_html_files_by_toc(zp, toc_fp, html_files)
            except:
                if self.VERBOSE:
                    print("No table of contents. Use file numbers for order.")
                html_files = html_files
            if self.VERBOSE:
                print("Table of contents of the Epub file:")
                for file in html_files:
                    print("    ", file)
                print("End of table of contents")

            # Extract the text from the different html files in the epub archive:

            new_text = ""
            for fp in html_files:
                fn = os.path.split(fp)[-1]
                data = zp.read(fp)
                text = codecs.decode(data, "utf-8")
                text = re.sub(r"<br />", "\n", text)
                soup = BeautifulSoup(text)
                root_tag = eval(self.root_tag_string)
                self.add_unique_tags(root_tag, src_fp, fn)
                text = self.convert_html2md(root_tag)
                new_text += "\n\n\n\n" + text
            new_text = re.sub("\n{3,}", "\n\n", new_text)

            return new_text


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    print("Testing finished.")
    
##    c = GenericEpubConverter()
##    c.make_dest_fp("a/b/c.epub")
##    print(c.dest_fp)
##    c = GenericEpubConverter("test/converted")
##    c.make_dest_fp("a/b/c.epub")
##    print(c.dest_fp)
##
##
##    input()

##    gen_converter = GenericEpubConverter("test/converted")
##    fp = r"test\Houellebecq 2019 - serotonine.epub"
##    gen_converter.convert_file(fp)
##    print("converted Houellebecq")
##    gen_converter.VERBOSE = True
##    gen_converter.convert_files_in_folder("test", ["epub",])
##    print("all files in test folder converted")
##
##    print("-"*60)

    import html2md_hindawi
    HindawiConverter = GenericEpubConverter("test/converted")
    HindawiConverter.convert_html2md = html2md_hindawi.markdownify
    HindawiConverter.toc_fn = "nav.xhtml"
    fp = r"test\26362727.epub"
    HindawiConverter.convert_file(fp)
    print("converted Hindawi text")







import re
import math
import random
import urllib.request as url
import os


if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)
from openiti.helper import ara

splitter = "#META#Header#End#"
ar_ch = "[ذ١٢٣٤٥٦٧٨٩٠ّـضصثقفغعهخحجدًٌَُلإإشسيبلاتنمكطٍِلأأـئءؤرلاىةوزظْلآآ]"
milestone = "Milestone300"
thresh = 1000


def text_cleaner(text):
    text = ara.normalize_ara_extra_light(text)
    text = re.sub("\W|\d|[A-z]", " ", text)
    text = re.sub(" +", " ", text)
    return text


def roundup(x, par):
    new_x = int(math.ceil(int(x) / float(par)) * par)
    return new_x


def generate_ids_through_permutations(char_string_for_ids, id_len_char, limit):
    list_char = list(char_string_for_ids) * id_len_char

    dic = {}
    iterations = 0

    while len(dic) < limit:
        new = "".join(random.sample(list_char, id_len_char))
        dic[new] = 0

        iterations += 1
        if iterations % 100000 == 0:
            print("\tITERATIONS: %d; DICTIONARY: %d" % (iterations, len(dic)))

    ids = "\n".join(list(dic.keys()))

    # where `L` is the length of IDs, `T` is the total number of unique IDs.
    file_name = "IDs_ASCII_L%d_T%d.txt" % (id_len_char, len(dic))
    with open(file_name, 'w', encoding='utf8') as outfile:
        outfile.write(ids)

    print("=" * 80)
    print("Generating TXT file with unique IDs, based on:")
    print("\tPermuations (L=%d) of: %s" % (id_len_char, char_string_for_ids))
    print("\tTotal number of IDs: %s" % '{:,}'.format(len(dic)))
    print("=" * 80)


def ar_ch_len(fp):
    """Count the length of a text in Arabic characters, given its URL"""

    try:
        with url.urlopen(fp) as f:
            book = f.read().decode('utf-8')
    except:
        with open(fp, mode="r", encoding="utf-8") as f:
            book = f.read()

        # splitter/header test
        if splitter in book:
            # split the header and body of the text:
            text = book.split(splitter)[1]

            # remove Editorial sections:
            text = re.sub("### \|EDITOR.+?### ", "### ", text,
                          flags = re.DOTALL)

            # count the number of Arabic letters or numbers:
            toks = re.findall(ar_ch, text)
            ar_ch_cnt = len([c for c in toks if c in ar_ch])
            print("{} Arabic character count: {}".format(fp, ar_ch_cnt))
            return ar_ch_cnt
        else:
            print("{} is missing the splitter!".format(fp))
            return 0


def read_header(fp):
    """Read only the OpenITI header of a file without opening the entire file.

    Args:
        fp (str): path to the text file

    Returns:
        header (list): A list of all metadata lines in the header
    """
    with open(fp, mode="r", encoding="utf-8") as file:
        header = []
        line = file.readline()
        i=0
        while i < 100:
        #while "#META#Header#End" not in line and i < 100:
            #if "#META#" in line or "#NewRec#" in line:
            header.append(line)
            if "#META#Header#End" in line:
                return header
            line = file.readline() # move to next line
            i += 1
    return header


def absolute_path(path):
    return os.path.abspath(path)



if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)

from openiti.helper import ara



# Moves all ara1 () file from all directories into a single directory

import os
import re
import shutil
import sys


def move_ara_files(source_dir, dest_dir):
    for root, dirs, files in os.walk(source_dir):
        # print("root: ",root)
        dirs[:] = [d for d in dirs]
        for file in files:
            if re.search("^\d{4}\w+\.\w+\.\w+-\w{4}(\.(mARkdown|inProgress|completed))+$", file):
            # if re.search("^\d{4}\w+\.\w+\.\w+-ara\d$", file):
                # to rename the files and keep only the text ID
                old_name = os.path.join(root, file)
                new_name = os.path.join(dest_dir, file.rsplit(".")[-1])
                shutil.copy(old_name, new_name)
                # copy th file with the original name
                # shutil.copy(os.path.join(root, file), dest_dir)


if __name__ == '__main__':

    source = input("Enter the source directory: ")
    target = input("Enter the target directory: ")

    if len(sys.argv) > 0:
        if not os.path.exists(source):
            print("source directory doesn't exists!")
        elif not os.path.exists(target):
            if target:
                os.makedirs(target)
                move_ara_files(source, target)
            else:
                print("please enter a target path")
        else:
            move_ara_files(source, target)
    else:
        print("give the path to the script...!")




"""
Replace markdown tags by html tags.

TO DO:
* replace ### \|+ headers not by <hx> but by <div><hx>title</hx>content</div>?
"""
import re
from bs4 import BeautifulSoup
from openiti.helper.templates import HTML_HEADER, HTML_FOOTER

##HTML_HEADER = """<html>
##<head>
##<style>
##  .entry-title {
##      font-size: 20px;
##  }
##</style>
##</head>
##<body>"""
##
##HTML_FOOTER = "</body></html>"

def header(text):
    return re.sub(".+?#META#HEADER#END", HTML_HEADER, text)

def unwrap(text):
    return re.sub("\n~~", " ", text)


def paragraphs(text):
    regex = r"\n# (.+?)(?=\n#|\n<p|\n<h|\n?</div>|\n\||$)"
    return re.sub(regex, r"\n<p>\1</p>", text, flags=re.DOTALL)


def poetry(text):
    p = r'<p class="verse"><span class="hemistych1">\1</span>'
    p += r' <span class="hemistych2">\2</span></p>'
    #text = re.sub("# (.+?) %~% (.+)", p, text)
    text = re.sub("<p>(.+?) %~% (.+?)</p>", p, text)
    return text


def editorials(text):
    r = r'<div class="editorial">\n\1\n</div>'
    return re.sub(r"### \|EDITOR\| ?\n(.+?)(?=$|\n###)", r, text,
                  flags=re.DOTALL)

def titles(text):
    def title_tag(match):
        level = len(match.group(1))
        cont = match.group(2)
        return '<h{}>{}</h{}>'.format(level, cont, level)
    return re.sub("### (\|+) (.+)", title_tag, text)

def dict_units(text):
    """Replace dictionary units mARkdown tags with html tags.

    Examples:
        >>> import md2html
        >>> md2html.dict_units("### $DIC_NIS$ Name of the entry")
        '<div class="entry descr-name"><span class="entry-title">Name of the entry</span>\\n</div>\\n'
        >>> md2html.dict_units("### $DIC_TOP$ Name of the place")
        '<div class="entry toponym"><span class="entry-title">Name of the place</span>\\n</div>\\n'
        >>> md2html.dict_units("### $DIC_LEX$ Word")
        '<div class="entry lexical"><span class="entry-title">Word</span>\\n</div>\\n'
        >>> md2html.dict_units("### $DIC_BIB$ Book title")
        '<div class="entry book"><span class="entry-title">Book title</span>\\n</div>\\n'
        >>> md2html.dict_units("### $BIO_MAN$ Name of the man")
        '<div class="entry man"><span class="entry-title">Name of the man</span>\\n</div>\\n'
        >>> md2html.dict_units("### $BIO_WOM$ Name of the woman")
        '<div class="entry woman"><span class="entry-title">Name of the woman</span>\\n</div>\\n'
        >>> md2html.dict_units("### $BIO_REF$ Cross-reference to a person")
        '<div class="entry cross-ref"><span class="entry-title">Cross-reference to a person</span>\\n</div>\\n'
        >>> md2html.dict_units("### $BIO_NLI$ List of names")
        '<div class="entry name-list"><span class="entry-title">List of names</span>\\n</div>\\n'
        >>> md2html.dict_units("### $CHR_EVE$ Event description")
        '<div class="entry event"><span class="entry-title">Event description</span>\\n</div>\\n'
        >>> md2html.dict_units("### $CHR_RAW$ Events description")
        '<div class="entry events-batch"><span class="entry-title">Events description</span>\\n</div>\\n'
        >>> md2html.dict_units("### $ Name of the man")
        '<div class="entry man"><span class="entry-title">Name of the man</span>\\n</div>\\n'
        >>> md2html.dict_units("### $$ Name of the woman")
        '<div class="entry woman"><span class="entry-title">Name of the woman</span>\\n</div>\\n'
        >>> md2html.dict_units("### $$$ Cross-reference to a person")
        '<div class="entry cross-ref"><span class="entry-title">Cross-reference to a person</span>\\n</div>\\n'
        >>> md2html.dict_units("### $$$$ List of names")
        '<div class="entry name-list"><span class="entry-title">List of names</span>\\n</div>\\n'
    """
    def dict_tag(match):
        tag_dict = {"WOM": "woman",
                    "MAN": "man",
                    "REF": "cross-ref",
                    "NLI": "name-list",
                    "EVE": "event",
                    "RAW": "events-batch",
                    "NIS": "descr-name",
                    "TOP": "toponym",
                    "LEX": "lexical",
                    "BIB": "book"}
        tag = match.group(1).strip()
        title = match.group(2).strip()
        cont = match.group(3).strip()
        if tag in tag_dict:
            tag = tag_dict[tag]
        else:
            tag = tag.lower()
        if title:
            title_span_tag = '<span class="entry-title">'
            if cont:
                fmt = '<div class="entry {}">{}{}</span>\n{}\n</div>\n'
                return fmt.format(tag, title_span_tag, title, cont)
            else:
                fmt = '<div class="entry {}">{}{}</span>\n</div>\n'
                return fmt.format(tag, title_span_tag, title)
        else:
            fmt = '<div class="entry {}">{}\n</div>\n'
            return fmt.format(tag, cont)

    def bio_tag(match):
        type_index = len(match.group(1))-1
        types = "man woman cross-ref name-list".split()
        title = match.group(2).strip()
        cont = match.group(3).strip()
        if cont:
            if title:
                fmt = '<div class="entry {}"><span class="entry-title">{}</span>\n{}\n</div>\n'
                return fmt.format(types[type_index], title, cont)
            else:
                fmt = '<div class="entry {}">{}\n</div>\n'
                return fmt.format(types[type_index], cont)
        else:
            fmt = '<div class="entry {}"><span class="entry-title">{}</span>\n</div>\n'
            return fmt.format(types[type_index], title)

    #regex = r"### \$[A-Z]{3}_([A-Z]{3})\$([^\n]*\n)(.*?)(?=###|<div>|$)"
    tag = "### \$[A-Z]{3}_([A-Z]{3})\$"
    title = "( ?[^\n]*)"
    cont = "((?:\n.*?)?)(?=###|<div|$)"
    regex = tag + title + cont
    text = re.sub(regex, dict_tag, text, flags=re.DOTALL)
##    print(text)
##    input()

#    regex = "### (\$+) ((?:.+?)?)\n((?:.+?)?)(?=\n###|\n<div>|$)"
    tag = "### (\$+)"
    title = "( ?[^\n]*)"
    cont = "((?:\n.*?)?)(?=###|<div|$)"
    #print(re.findall(tag+title+cont, text, flags=re.DOTALL))
    regex = tag + title + cont
##    for x in re.findall(regex, text, flags=re.DOTALL):
##        print(x)
##    input()
    text = re.sub(regex, bio_tag, text, flags=re.DOTALL)
    return text

def events(text):
    def event_tag(match):
        title = match.group(2).strip()
        cont = match.group(3).strip()
        if match.group(1):
            event_class = "events-batch"
        else:
            event_class = "event"
        if title:
            if cont:
                fmt = '<div class="{}"><span class="entry-title">{}</span>{}\n</div>\n'
                return fmt.format(event_class, title, cont)
            else:
                fmt = '<div class="{}"><span class="entry-title">{}</span>\n</div>\n'
                return fmt.format(event_class, title)
        else:
            if cont:
                fmt = '<div class="{}">{}\n</div>\n'
                return fmt.format(event_class, cont)
            else:
                return '<div class="{}">\n</div>\n'.format(event_class, title)

    regex = "### @ ((?:RAW)?) ?(.+?)(?=\n###|$|\n<div)"
    tag = "### @ ((?:RAW)?)"
    title = "( ?[^\n]*)"
    cont = "((?:\n.*?)?)(?=###|<div|$)"
    #print(re.findall(tag+title+cont, text, flags=re.DOTALL))
    regex = tag + title + cont
    
    return re.sub(regex, event_tag, text, flags=re.DOTALL)

def tables(text):
    """Replace markdown table tags with html table tags.

    Examples:
        >>> text = "# first paragraph\\n\
| table header 1 | header 2 |\\n\
|:---------------|----------|\\n\
| table row 1 col 1 | table row 1 col 2|\\n\
| table row 2 col 1 | table row 2 col 2|"
        >>> import md2html
        >>> md2html.tables(text)
        '# first paragraph\\n\
<table>\\n\
<tr>\\n<th>table header 1</th><th>header 2</th>\\n</tr>\\n\
<tr>\\n<td>table row 1 col 1</td><td>table row 1 col 2</td>\\n</tr>\\n\
<tr>\\n<td>table row 2 col 1</td><td>table row 2 col 2</td>\\n</tr>\\n\
</table>\\n'
        """
    def format_table(match):
        table = ""
        rows = match.group(1).strip().splitlines()
        #print(rows)
        if re.search("\A\| *[:\-]+", rows[1]): # if a header row is underlined
            header = rows.pop(0)
            del rows[0] # remove the underlining row
            # remove the starting and ending pipes, and wrap each cell in html tags:
            header = re.sub("\A\||\| *\Z", "", header)
            header = ["<th>" + x.strip() + "</th>" for x in header.split("|")]
            table += "<tr>\n{}\n</tr>\n".format("".join(header))
        for row in rows:
            row = re.sub("\A\||\| *\Z", "", row)
            row = ["<td>" + x.strip() + "</td>" for x in row.split("|")]
            table += "<tr>\n{}\n</tr>\n".format("".join(row))
        return "\n<table>\n{}</table>\n".format(table)
    
    text = re.sub("\n(\|.+?)(?:\n(?![~|])|$)", format_table, text, flags=re.DOTALL)  
    return text

def pages(text):
    def page_tag(match):
        return '<a class="page" n="{}.{}"></a>'.format(group(1), group(2))
    return re.sub("PageV(\d+)P(\d+)", page_tag, text)

def milestones(text):
    return re.sub("Milestone(\d+)", '<a class="milestone" n="\1">', text)

def replace_by_regex(text):
    text = header(text)
    text = unwrap(text)
    text = paragraphs(text)
    text = editorials(text)
    text = poetry(text)
    text = titles(text)
    text = dict_units(text)
    text = events(text)
    text = tables(text)
    text = pages(text)
    text = milestones(text)
    return text


def replace_by_split(text):
    # split the text on (combinations) of new lines, spaces or html tags,
    # keeping the split characters:
    spl = re.split("((?:\n| |<[^>]+>)+)", text)
    text = ""
    skip_until = 0
    for i, t in enumerate(spl):
        if i < skip_until:
            pass
        elif t.startswith("@"): # named entities
            cls, prefix, n = re.findall("@([A-Za-z@]+)(\d)(\d+)", t)[0]
            cls = " ".join(cls.split("@"))
            cont = "".join(spl[i+2:i+1+(2*int(n))])[int(prefix):]
            prefix = spl[i+2][:int(prefix)]
            fmt = '{}<span class="named-ent {}">{}</span>'
            text += fmt.format(prefix, cls, cont)
            skip_until = i + 1 + (int(n)*2)
        elif t.startswith("Y"):
            text += '<a class="{}" n="{}"></a>'.format(t[:2], t[2:])

        else:
            text += t
    return text


def convert(text):
    text = replace_by_regex(text)
    text = replace_by_split(text)
    return HTML_HEADER + text + HTML_FOOTER


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    text = """
### |EDITOR|
# Editorial line 1
# Editorial line 2
# Editorial line 3
### | Heading 1
### || Heading 2
### $$$ Dict unit 1
# paragraph 1 paragraph1
~~paragraph 1 continued
### $DIC_TOP$ Dict unit 2
# poem line 1 hemi 1 %~% poem line 1 hemi2
# paragraph 2 @TOP23 biMadinat al Salam
### $DIC_LEX$
# untitled dict unit
~~unnamed dict unit continued
### @ Event 1
### @ RAW
# Event 1
# Event 2 @PER03 Abdallah ibn Ahmad @SRC02 Kitab al-Buldan
### $$ Name of the woman
# Description of the woman's life
| table header 1 | header 2 |
|:---------------|----------|
| table row 1 col 1 | table row 1 col 2|
| table row 2 col 1 | table row 2
~~col 2|
### |EDITOR|
# Editorial 2 line 1
# Editorial 2 line 2
"""
    html = convert(text)
    #print(html.strip())
    print(BeautifulSoup(html, 'html.parser').prettify())



"""Get selected annotation issues from GitHub and print them
or save them as a tsv file.


Usage example:
    issues = get_issues("OpenITI/Annotation",
                        #issue_labels=["in progress"],
                        state="all"
                        )
    issues = define_text_uris(issues)
    uri_dict = sort_issues_by_uri(issues)
    print_issues_by_uri(uri_dict, "test.tsv")
"""

from github import Github
import re

URI_REGEX = r"\d{4}[A-Z][a-zA-Z]+(?:\.[A-Z][a-zA-Z]+)?(?:\.\w+-[a-z]{3}\d+)?"


def define_text_uris(issues, verbose=False):
    """Define which text uri the issue pertains to.
    Store the uri in the issue object (issue.uri).

    Args:
        issues (list): a list of github issue objects.
        verbose (bool): if verbose, print issues for which no uris were found.

    Returns:
        issues (list): the list of updated github issue objects
    """
    for issue in issues:
        if re.findall(URI_REGEX, issue.title):
            issue.uri = re.findall(URI_REGEX, issue.title)[0]
        elif re.findall("OLD URI.+?"+URI_REGEX, issue.body):
            issue.uri = re.findall("OLD URI.+?"+URI_REGEX, issue.body)[0]
        elif re.findall(URI_REGEX, issue.body):
            issue.uri = re.findall(URI_REGEX, issue.body)[0]
        elif issue.comments:
            for c in issue.get_comments():
                if re.findall(URI_REGEX, c.body):
                    issue.uri = re.findall(URI_REGEX, c.body)[0]
                    break
                issue.uri = ""
        else:
            issue.uri = ""
            if verbose:
                print("no uri found")
                print("    number:", issue.number)
                print("    title:", issue.title)
                print("    body:", issue.body)
                print("    comments:", [x.body for x in issue.get_comments()])
                input("Press enter to continue")
        #print("issue.uri:", issue.uri)
    return issues


def sort_issues_by_uri(issues):
    """Create a dictionary with the uris as keys.

    Args:
        issues (list): a list of github issue objects

    Returns:
        uri_dict (dict): a dictionary:
            key: uri
            value: list of github issue objects related to this uri
    """
    uri_dict = dict()
    for issue in issues:
        if issue.uri: # leave out issues for which we didn't find a URI
            if issue.uri in uri_dict:
                uri_dict[issue.uri].append(issue)
            else:
                uri_dict[issue.uri] = [issue]
    return uri_dict

def print_issues_by_uri(uri_dict, save_fp=""):
    """Print a tab-delimited list of uris with the issues connected to them.
    URI     Issue number    Issue label

    Args:
        uri_dict (dict): key: uri,
                         value: list of github issues related to this uri

    Returns:
        None
    """
    s = "{}\t{}\t{}\t{}\n".format("URI", "Issue number",
                                  "Issue label", "Issue state")
    for uri, issues in sorted(uri_dict.items()):
        for issue in issues:
            for label in issue.labels:
                s += "{}\t{}\t{}\t{}\n".format(uri, issue.number,
                                               label.name, issue.state)
    
    if save_fp:
        print("Saved list of issues to", save_fp)
        with open(save_fp, mode="w", encoding="utf-8") as file:
            file.write(s)
    else:
        print(s)


def get_issues(repo_name, user=None, password=None,
               issue_labels=None, state="open"):
    """Get all issues connected to a specific github repository.

    Args:
        repo_name (str): the name of the Github repository
        user (str): username
        password (str): password
        issue_labels (list; default: None): a list of github issue label names;
            only the issues with an issue label name in this
            list will be downloaded;
            if set to None, all issues will be downloaded
        state (str; default: "open"): only the issues with this state
            (open/closed/all) will be downloaded.

    Returns:
        issues (list): a list of github issues. 
    """
    if user == None:
        user = input("User: ")
    if password == None:
        password = input("Password: ")

    g = Github(user, password)
    print("logged in")
    repo = g.get_repo(repo_name)
    print("issue labels in this repo:")
    for label in repo.get_labels():
        print("  ", label)
    print("getting relevant issues from GitHub...")
    if state:
        issues = repo.get_issues(state=state)
    else:
        issues = repo.get_issues(state="all")
    if issue_labels != None:
        filtered = []
        for i in issues:
            for il in issue_labels:
                if il in [label.name for label in i.labels]:
                    filtered.append(i)
                    break
        return filtered
    else:
        return issues


if __name__ == "__main__":
    issues = get_issues("OpenITI/Annotation",
                        #issue_labels=["in progress"],
                        state="all"
                        )
    issues = define_text_uris(issues)
    uri_dict = sort_issues_by_uri(issues)
    print_issues_by_uri(uri_dict, "test.tsv")
    


        



import re


def deNoise(text):
    """Remove non-consonantal characters from Arabic text.

    Examples:
        >>> deNoise("وَالَّذِينَ يُؤْمِنُونَ بِمَا أُنْزِلَ إِلَيْكَ وَمَا أُنْزِلَ مِنْ قَبْلِكَ وَبِالْآخِرَةِ هُمْ يُوقِنُونَ")
        'والذين يؤمنون بما أنزل إليك وما أنزل من قبلك وبالآخرة هم يوقنون'
        >>> deNoise(" ْ ً ٌ ٍ َ ُ ِ ّ ۡ ࣰ ࣱ ࣲ ٰ ")
        '              '
    """
    noise = re.compile(""" ّ    | # Tashdīd / Shadda
                           َ    | # Fatḥa
                           ً    | # Tanwīn Fatḥ / Fatḥatān
                           ُ    | # Ḍamma
                           ٌ    | # Tanwīn Ḍamm / Ḍammatān
                           ِ    | # Kasra
                           ٍ    | # Tanwīn Kasr / Kasratān
                           ْ    | # Sukūn
                           ۡ    | # Quranic Sukūn
                           ࣰ    | # Quranic Open Fatḥatān
                           ࣱ    | # Quranic Open Ḍammatān
                           ࣲ    | # Quranic Open Kasratān
                           ٰ    | # Dagger Alif
                           ـ     # Taṭwīl / Kashīda
                       """, re.VERBOSE)
    return re.sub(noise, "", text)


def normalize(text, replacement_tuples=[]):
    """Normalize Arabic text by replacing complex characters by simple ones.

    Args:
        text (str): the string that needs to be normalized
        replacement_tuples (list of tuple pairs): (character, replacement)

    Examples:
        >>> normalize('AlphaBet', [("A", "a"), ("B", "b")])
        'alphabet'
    """
    for char, repl in replacement_tuples:
        text = text.replace(char, repl)
    return text


def normalize_ara_light(text):
    """Lighlty normalize Arabic strings:
    fixing only Alifs, Alif Maqsuras;
    replacing hamzas on carriers with standalone hamzas

    Examples:
        >>> normalize_ara_light("ألف الف إلف آلف ٱلف")
        'الف الف الف الف الف'
        >>> normalize_ara_light("يحيى")
        'يحيي'
        >>> normalize_ara_light("مقرئ فيء")
        'مقرء فء'
        >>> normalize_ara_light("قهوة")
        'قهوة'
    """
    repl = [("أ", "ا"), ("ٱ", "ا"), ("آ", "ا"), ("إ", "ا"),    # alifs
            ("ى", "ي"),                                        # alif maqsura
            ("يء", "ء"), ("ىء", "ء"), ("ؤ", "ء"), ("ئ", "ء"),  # hamzas
            ]
    return normalize(text, repl)
    

def normalize_ara_heavy(text):
    """Normalize Arabic text by simplifying complex characters:
    alifs, alif maqsura, hamzas, ta marbutas

    Examples:
        >>> normalize_ara_heavy("ألف الف إلف آلف ٱلف")
        'الف الف الف الف الف'
        >>> normalize_ara_heavy("يحيى")
        'يحيي'
        >>> normalize_ara_heavy("مقرئ فيء")
        'مقر في'
        >>> normalize_ara_heavy("قهوة")
        'قهوه'
    """
    repl = [("أ", "ا"), ("ٱ", "ا"), ("آ", "ا"), ("إ", "ا"),  # alifs
            ("ى", "ي"),                                      # alif maqsura
            ("ؤ", ""), ("ئ", ""), ("ء", ""),                 # hamzas
            ("ة", "ه")                                       # ta marbuta
            ]
    return normalize(text, repl)


def denormalize(text):
    """Replace complex characters with a regex covering all variants.

    Examples:
        >>> denormalize("يحيى")
        'يحي[يى]'
        >>> denormalize("هوية")
        'هوي[هة]'
        >>> denormalize("مقرئ")
        'مقر(?:[ؤئ]|[وي]ء)'
        >>> denormalize("فيء")
        'في(?:[ؤئ]|[وي]ء)'
    """
    alifs = '[إأٱآا]'
    alif_maqsura = '[يى]\\b'
    alif_maqsura_reg = '[يى]'
    ta_marbuta = '[هة]\\b'
    ta_marbuta_reg = '[هة]'
    hamzas = '[ؤئء]'
    hamzas_reg = '(?:{}|{}ء)'.format('[ؤئ]', '[وي]')
    # Applying deNormalization
    text = re.sub(alifs, alifs, text)
    text = re.sub(alif_maqsura, alif_maqsura_reg, text)
    text = re.sub(ta_marbuta, ta_marbuta_reg, text)
    text = re.sub(hamzas, hamzas_reg, text)
    return text


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    



"""A generic converter for text into OpenITI format.

Subclass the GenericConverter to make a converter for a new
source of scraped books.

TO DO:
    import the deNoise function from the library instead of
        defining it here in the module.
"""

import os
import re
import textwrap


def deNoise(text):
    """Eliminate all diacritics from the text."""
    noise = re.compile(""" َ  | # Fatha
                             ً  | # Tanwin Fath
                             ُ  | # Damma
                             ٌ  | # Tanwin Damm
                             ِ  | # Kasra
                             ٍ  | # Tanwin Kasr
                             ْ  | # Sukun
                             ـ | # Tatwil/Kashida
                             ّ # Tashdid
                             """, re.VERBOSE)
    text = re.sub(noise, "", text)
    text = re.sub("ﭐ", "ا", text) # replace alif-wasla with simple alif
    return text

class GenericConverter(object):

    def __init__(self):
        """Initialize the default values."""
        self.magic_value = "######OpenITI#\n\n\n"
        self.header_splitter = "\n\n#META#Header#End#\n"
        self.endnote_splitter = "\n\n### |EDITOR|\nENDNOTES:\n\n"
        self.dest_folder = "converted"
        self.extension = ".automARkdown"

        # settings for the chunking of the text:

        self.chunk_length = 300
        self.milestone_tag = "Milestone300"

        # regexes describing Arabic characters and tokens:

        self.ara_char = "[؀-ۿࢠ-ࣿﭐ-﷿ﹰ-﻿]"
        ara_tok = "{}+|[^{}+".format(self.ara_char, self.ara_char[1:])
        self.ara_tok = re.compile(ara_tok)

        self.VERBOSE = False

##        self.conversion_procedure(source_fp, dest_dir)


    def convert_files_in_folder(self, source_folder, extensions=[]):
        """Convert all files in a folder to OpenITI format.

        Args:
            source_folder (str): path to the folder that contains
                the files that must be converted.
            extensions (list): list of extensions; if this list is not empty,
                only files with an extension in the list should be converted.

        Returns:
            None
        """
        for source_fn in os.listdir(source_folder):
            if extensions:

                # make sure extensions are preceded by a period:

                extensions = [re.sub("\.+", ".", "."+ext) for ext in extensions]

                # convert a file only if its extension is in the list:

                if os.path.splitext(source_fn)[-1] in extensions:
                    self.convert_file(os.path.join(source_folder, source_fn))
            else:
                self.convert_file(os.path.join(source_folder, source_fn))


    def convert_file(self, source_fp):
        """Convert one file to OpenITI format.

        Args:
            source_fp (str): path to the file that must be converted.

        Returns:
            None
        """
        if self.VERBOSE:
            print("converting", source_fp)
        dest_fp = self.make_dest_fp(source_fp)
        metadata = self.get_metadata(source_fp)

        text = self.get_data(source_fp)
        #print(1, text)
        text = self.pre_process(text)
        #print(2, text)

        text = self.add_page_numbers(text)
        #print(3, text)
        text = self.add_structural_annotations(text)
        #print(4, text)
        text, notes = self.remove_notes(text)
        #print(5, text)
        text = self.reflow(text)
        #print(6, text)
        text = self.add_milestones(text)
        #print(7, text)

        text = self.post_process(text)
        #print(8, text)

        text = self.compose(metadata, text, notes)
        #print(9, text)

        self.save_file(text, dest_fp)
        #print("saved text to", dest_fp)


    def make_dest_fp(self, source_fp):
        """Make a filepath for the converted text file.

        ### Overwrite this method in subclass ###

        Args:
            source_fp (str): path to the file that must be converted.

        Returns:
            dest_fp (str): path to the converted text file.
        """
##        print("""WRITE A make_dest_fp FUNCTION \
##THAT CREATES A FILEPATH FOR THE CONVERTED TEXT FILE.""")
        source_folder, source_fn = os.path.split(source_fp)
        source_fn = os.path.splitext(source_fn)[0]+self.extension
        if self.dest_folder == "converted":
            self.dest_folder = os.path.join(source_folder, "converted")
        self.dest_fp = os.path.join(self.dest_folder, source_fn)
        return self.dest_fp


    def get_metadata(self, source_fp):
        """Gets the metadata from the file at source_fp.

        ### Overwrite this method in subclass ###

        Args:
            source_fp (str): path to the original file

        Returns:
            metadata (str): metadata formatted in OpenITI format
                (including the magic value and the header splitter)
        """
        metadata = ""
##        print("""WRITE A get_metadata FUNCTION \
##THAT COMPOSES THE METADATA HEADER FOR THE TEXT FILE.""")
        metadata = self.magic_value + metadata + self.header_splitter
        return metadata


    def get_data(self, source_fp):
        """Extract the data from the source_fp and format it in OpenITI format.

        ### Overwrite this method in subclass ###

        Args:
            source_fp (str): path to the original file

        Returns:
            text (str): the text in its initial state
        """
##        print("""WRITE A get_data FUNCTION \
##THAT GETS THE TEXT FROM THE source_fp""")
        with open(source_fp, mode="r", encoding="utf-8") as file:
            text = file.read()
        return text


    def pre_process(self, text):
        """Remove unwanted features from the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its initial state

        Returns:
            text (str): pre-processed text
        """
        text = deNoise(text)
        return text


    def add_page_numbers(self, text):
        """Add the page numbers in OpenITI format to the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state

        Returns:
            text_w_pages (str): text with the page numbers added
                (in OpenITI format).
        """
##        print("""WRITE AN add_page_numbers FUNCTION \
##THAT ADDS THE PAGE NUMBERS TO THE TEXT""")
        text_w_pages = text
        return text_w_pages


    def add_structural_annotations(self, text):
        """Add strucural annotations in OpenITI format to the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state

        Returns:
            text_w_pages (str): text with the structural markup
                (in OpenITI format) added.
        """
##        print("""WRITE AN add_structural_annotations FUNCTION \
##THAT ADDS THE PAGE NUMBERS TO THE TEXT""")
        ann_text = text
        return ann_text


    def remove_notes(self, text):
        """Remove the footnotes from the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str)

        Returns:
            text_without_notes (str): text from which the notes were removed.
            notes (str): the footnotes converted to endnotes
                (if there are endnotes: add the endnote_splitter before them)
        """
##        print("""WRITE A remove_notes FUNCTION \
##THAT SPLITS THE NOTES FROM THE TEXT""")
        text_without_notes = text
        notes = ""
        if notes:
            notes = self.endnote_splitter + notes
        return text_without_notes, notes


    def reflow(self, text, max_len=72):
        """wrap long lines, i.e. break long lines down
        by inserting a new line character (and two tildes)
        so that no line is longer than max_len characters.
        See https://docs.python.org/3.1/library/textwrap.html
        Mark the start of paragraphs with "# ".

        Args:
            text (str): text string to be wrapped
            max_len (int): "width" = max number of characters in a line

        Returns:
            newtext (str): wrapped string

        Examples:
            >>> import generic_converter
            >>> g = generic_converter.GenericConverter()
            >>> text = "# This is a short paragraph\\n\\n\
# This is a longer paragraph (longer than max_len)"
            >>> g.reflow(text, max_len=40)
            '# This is a short paragraph\\n\\n\
# This is a longer paragraph (longer\\n\
~~than max_len)'
            
        """
        newtext = []

        # split the text, keeping the number of newline characters:

        text = re.sub("\r", "\n", text)
        text = re.split("(\n\n+)", text)

        # wrap each line:

        ignoreTuple = ("#000000#", "#NewRec#", "#####", "### ",
                       "#META#", "Page")
        for line in text:
            if not line.startswith(ignoreTuple):
                if line != "" and "\n\n" not in line:

                    # make sure the new lines that signal the end of a tag
                    # are treated differently from textwrap newlines:
                    
                    sublines = line.split("\n")
                    new_line = []
                    for s in sublines:
                        new_line.append("\n~~".join(textwrap.wrap(s, max_len)))
                    line = "\n".join(new_line)
##                if line != "" and "\n" not in line:
##                    if not line.startswith("# "):
##                        line = "# " + line
##                    line = "\n~~".join(textwrap.wrap(line, max_len))
            newtext.append(line)

        # re-assemble the text:
        
        newtext = "".join(newtext)

        # unwrap lines that are less than 10 characters long:
        
        newtext = re.sub(r"[\r\n]+~~(.{1,6}?[\r\n]+)", r" \1", newtext)
        return newtext


    def add_milestones(self, text):
        """Add a milestone tag after a pre-defined number of words.

        The number of words is defined in the self.chunk_length variable.

        Example:
            >>> import generic_converter
            >>> g = generic_converter.GenericConverter()
            >>> g.chunk_length = 5
            >>> g.milestone_tag = "###"
            >>> text = '\
بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ\
 * الْحَمْدُ لِلَّهِ رَبِّ الْعَالَمِينَ\
 * الرَّحْمَنِ الرَّحِيمِ\
 * مَالِكِ يَوْمِ الدِّينِ\
 * إِيَّاكَ نَعْبُدُ وَإِيَّاكَ نَسْتَعِينُ\
 * اهْدِنَا الصِّرَاطَ الْمُسْتَقِيمَ\
 * صِرَاطَ الَّذِينَ أَنْعَمْتَ عَلَيْهِمْ غَيْرِ الْمَغْضُوبِ عَلَيْهِمْ وَلَا الضَّالِّينَ'
            >>> g.add_milestones(text)
            '\
بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ\
 * الْحَمْدُ ### لِلَّهِ رَبِّ الْعَالَمِينَ\
 * الرَّحْمَنِ الرَّحِيمِ\
 * ### مَالِكِ يَوْمِ الدِّينِ\
 * إِيَّاكَ نَعْبُدُ ### وَإِيَّاكَ نَسْتَعِينُ\
 * اهْدِنَا الصِّرَاطَ الْمُسْتَقِيمَ\
 * ### صِرَاطَ الَّذِينَ أَنْعَمْتَ عَلَيْهِمْ غَيْرِ ### الْمَغْضُوبِ عَلَيْهِمْ وَلَا الضَّالِّينَ'
        """
        # start counting only after the first ### | tag
        try:
            editorial, text = text.split("### | ", maxsplit=1)
            text = "### | " + text
        except:
            editorial = ""
        ms_text = ""  # the text with the milestones added
        token_count = 0
        milestones = 0
        for token in self.ara_tok.finditer(text):
            #print(token_count, [token.group()])
            if re.findall(self.ara_char, token.group()[0]):
                if token_count == self.chunk_length:
                    token_count = 1
                    milestones += 1
                    ms_text += self.milestone_tag
                else:
                    token_count += 1
            else:
                pass  # do not increment the token count for non-Arabic tokens
            ms_text += token.group()

        # add spaces before and after the milestone_tag:
        ms_text = re.sub(" *({}) *".format(self.milestone_tag),
                         r" \1 ", ms_text)

        return editorial + ms_text


    def post_process(self, text):
        """Carry out post-processing operations on the text.

        ### Overwrite this method in subclass ###

        Args:
            text (str): the text in its current conversion state

        Returns:
            processed (str): the text after post-processing operations.
        """
##        print("""WRITE A post_process FUNCTION \
##THAT CARRIES OUT THE LAST POST-PROCESSING OPERATIONS ON THE TEXT""")
        processed = re.sub("[\n\r]{2,}(?![#\|])", "\n\n# ", text)
        return processed

    def compose(self, metadata, text, notes):
        """Compose the final text from its parts.

        ### Overwrite this method in subclass ###

        Args:
            metadata (str): the metadata string in OpenITI format
                (including the magic_value and header_splitter)
            text (str): the main text (incl. page numbers and milestones)
            notes (str): the footnotes, converted to endnotes.

        Returns:
            composite_text (str)
        """
        composite_text = metadata + text + notes
        return composite_text


    def save_file(self, text, dest_fp):
        if not os.path.exists(os.path.split(dest_fp)[0]):
            os.makedirs(os.path.split(dest_fp)[0])
        with open(dest_fp, mode="w", encoding="utf-8") as file:
            file.write(text)


if __name__ == "__main__":
    import doctest
    doctest.testmod()

    #input("Press Enter to start converting")

    conv = GenericConverter()
    conv.convert_file("test/test.txt")







"""Convert html to Markdown.

This program is an adaptation of python-markdownify
(https://github.com/matthewwithanm/python-markdownify)
to output OpenITI mARkdown.
It also adds methods for tables and images,
and a post-processing method. 

You can use this class as a base class and subclass it
to add methods, adapt the post-processing method etc.

E.g.:
    def Class Hindawi_converter(html2md.MarkdownConverter):   
        def post_process_md(self, text):
            text = super().post_process_md(text)
            # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
            text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
            # replace placeholders for spaces in tables: 
            text = re.sub("ç", " ", text)
            return text


Examples (doctests):

    Headings: h1

    >>> import html2md
    >>> h = '<h1>abc</h1>'
    >>> html2md.markdownify(h)
    '\\n\\n### | abc\\n\\n'

    NB: heading style is OpenITI mARkdown style by default,
        but can be set to other styles as well:

    >>> h = '<h1>abc</h1>'
    >>> html2md.markdownify(h, md_style=UNDERLINED)
    '\\n\\nabc\\n===\\n\\n'

    >>> h = '<h1>abc</h1>'
    >>> html2md.markdownify(h, md_style=ATX)
    '\\n\\n# abc\\n\\n'

    Paragraphs (<p>):
    
    >>> h = "<p>abc</p>"
    >>> html2md.markdownify(h)
    '\\n\\n# abc\\n\\n'

    >>> h = "<p>abc</p>"
    >>> html2md.markdownify(h, md_style=ATX)
    '\\n\\nabc\\n\\n'


    Divs without class or with an unsupported class are stripped:
    
    >>> h = 'abc\
             <div>def</div>\
             ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'
        
    >>> h = 'abc\
             <div class="unknown_div_class">def</div>\
             ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'


    Spans without class or with an unsupported class are stripped:
    
    >>> h = 'abc <span>def</span> ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'
    
    >>> h = 'abc <span class="unknown_span_class">def</span> ghi'
    >>> html2md.markdownify(h)
    'abc def ghi'


    Links: 

    >>> h = '<a href="a/b/c">abc</a>'
    >>> html2md.markdownify(h)
    '[abc](a/b/c)'

    
    Unordered lists: 

    >>> h = '<ul><li>item1</li><li>item2</li></ul>'
    >>> html2md.markdownify(h)
    '\\n* item1\\n* item2\\n\\n'
    
    Ordered lists:

    >>> h = '<ol><li>item1</li><li>item2</li></ol>'
    >>> html2md.markdownify(h)
    '\\n1. item1\\n2. item2\\n\\n'

    Nested lists:
    
    >>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
    >>> html2md.markdownify(h)
    '\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'

    Italics (<i> and <em> tags):

    >>> h = 'abc <em>def</em> ghi'
    >>> html2md.markdownify(h)
    'abc *def* ghi'

    >>> h = 'abc <i>def</i> ghi'
    >>> html2md.markdownify(h)
    'abc *def* ghi'


    Bold (<b> and <strong> tags):

    >>> h = 'abc <b>def</b> ghi'
    >>> html2md.markdownify(h)
    'abc **def** ghi'

    >>> h = 'abc <strong>def</strong> ghi'
    >>> html2md.markdownify(h)
    'abc **def** ghi'

    Tables:
    
    >>> h = '\
    <table>\
      <tr>\
        <th>th1aaa</th><th>th2</th>\
      </tr>\
      <tr>\
        <td>td1</td><td>td2</td>\
      </tr>\
    </table>'
    >>> html2md.markdownify(h)
    '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'
            
    # i.e.,
    # | th1aaa | th2 |
    # | td1    | td2 |
"""

from bs4 import BeautifulSoup, NavigableString
import re
import six


convert_heading_re = re.compile(r'convert_h(\d+)')
line_beginning_re = re.compile(r'^', re.MULTILINE)
whitespace_re = re.compile(r'[\r\n\s\t ]+')
FRAGMENT_ID = '__MARKDOWNIFY_WRAPPER__'
wrapped = '<div id="%s">%%s</div>' % FRAGMENT_ID



# Heading styles
ATX = 'atx'
ATX_CLOSED = 'atx_closed'
UNDERLINED = 'underlined'
OPENITI = 'OpenITI'
SETEXT = UNDERLINED


def escape(text):
##    if not text:
##        return ''
##    return text.replace('_', r'\_')
    return text


def _todict(obj):
    return dict((k, getattr(obj, k)) for k in dir(obj) if not k.startswith('_'))


class MarkdownConverter(object):
    class DefaultOptions:
        strip = None
        convert = None
        autolinks = True
        md_style = OPENITI
        bullets = '*+-'  # An iterable of bullet types.

    class Options(DefaultOptions):
        pass

    def __init__(self, **options):
        # Create an options dictionary. Use DefaultOptions as a base so that
        # it doesn't have to be extended.
        self.options = _todict(self.DefaultOptions)
        self.options.update(_todict(self.Options))
        self.options.update(options)
        if self.options['strip'] is not None and self.options['convert'] is not None:
            raise ValueError('You may specify either tags to strip or tags to'
                             ' convert, but not both.')

    def convert(self, html):
        """Convert html to markdown.

        # We want to take advantage of the html5 parsing, but we don't actually
        # want a full document. Therefore, we'll mark our fragment with an id,
        # create the document, and extract the element with the id.
        """
        
        html = wrapped % html
        soup = BeautifulSoup(html, 'html.parser')
        text = self.process_tag(soup.find(id=FRAGMENT_ID), children_only=True)

##        # post-processing: remove unneeded blank lines and spaces:
##        
##        # remove leading and trailing spaces in lines:
##        text = re.sub(r" *(\n+) *", r"\1", text)
##        # remove unwanted additional spaces and lines:
##        text = re.sub(r"\n{3,}", r"\n\n", text)
##        text = re.sub(r" +", r" ", text)
##        # remove blank lines marked with "DELETE_PREVIOUS_BLANKLINES" tag
##        text = re.sub(r"\n+DELETE_PREVIOUS_BLANKLINES", "", text)
##        # replace placeholders for spaces in tables: 
##        text = re.sub("ç", " ", text)
##        return text
        return self.post_process_md(text)

    def process_tag(self, node, children_only=False):
        """Process each tag and its children.
        """
        #print(node.name, node.attrs)
        text = ''

        # Convert the children first
        for el in node.children:
            
            if isinstance(el, NavigableString):
                text += self.process_text(six.text_type(el))
            else:
                text += self.process_tag(el)


        if not children_only:
            convert_fn = getattr(self, 'convert_%s' % node.name, None)
            if convert_fn and self.should_convert_tag(node.name):
                #print("text is now:")
                #input(text)
                text = convert_fn(node, text)
        return text

    def process_text(self, text):
        #return escape(whitespace_re.sub(' ', text or ''))
        if whitespace_re.sub('', text or ''):
            return whitespace_re.sub(' ', text)
        return ""

    def fill_out_columns(self, match):
        """Find the longest cell in a column; add spaces to shorter columns."""

        # split the table into a list of lists:
        
        table = match.group(1)
        #print(table)
        if "|--" in table:
            header = True
        else:
            header = False
        table = table.splitlines()
        if header:
            del table[1]  # remove horizontal lines separating header from body
        table = [(row.split("|"))[1:] for row in table]  # remove empty string

        # get length of longest cell per column:
        
        column_length = dict()
        for row in table:
            for i, cell in enumerate(row):
                try:
                    if len(cell) > column_length[i]:
                        column_length[i] = len(cell)
                except:
                    column_length[i] = len(cell)

        # fill out shorter cells:
        
        new_table = []
        for row in table:
            new_row = []
            for i, cell in enumerate(row):
                new_cell = cell + " "*(column_length[i]-len(cell))
                new_cell = new_cell + (" " * len(re.findall("لا|لأ|لإ", new_cell)))
                #combining = re.findall("[ؐ-ًؚ-ٰٟۖ-ۜ۟-۪ۤۧۨ-ۭ]", new_cell)
                #new_cell = re.sub(" {%s}" % len(combining), "", new_cell)
                new_row.append(new_cell)
            new_table.append(new_row)

        # re-build the table:
        
        new_table = [" | ".join(row).strip() for row in new_table]
        if header:
            lines = " | ".join(["-" * v for k,v in column_length.items()])
            new_table = [new_table[0], lines.strip()]+new_table[1:]
        new_table = "\n| ".join(new_table)
        new_table = "\n\n| {}\n\n".format(new_table)
        
        #print("new_table: ")
        #print(new_table)
        return new_table
        

    def post_process_md(self, text):
        # remove leading and trailing spaces in lines:
        text = re.sub(r" *(\n+) *", r"\1", text)
        # remove unwanted additional spaces and lines:
        text = re.sub(r"\n{3,}", r"\n\n", text)
        text = re.sub(r" +", r" ", text)
        # fill out columns in tables:
        text = re.sub("\n\n(\|.+?)\n\n", self.fill_out_columns, text, flags=re.DOTALL)
        return text

    def __getattr__(self, attr):
        # Handle headings
        m = convert_heading_re.match(attr)
        if m:
            n = int(m.group(1))

            def convert_tag(el, text):
                return self.convert_hn(n, el, text)

            convert_tag.__name__ = 'convert_h%s' % n
            setattr(self, convert_tag.__name__, convert_tag)
            return convert_tag

        raise AttributeError(attr)

    def should_convert_tag(self, tag):
        """Check whether a tag should be converted or simply stripped"""
        tag = tag.lower()
        strip = self.options['strip']
        convert = self.options['convert']
        if strip is not None:
            return tag not in strip
        elif convert is not None:
            return tag in convert
        else:
            return True

    def indent(self, text, level):
        """Add tab indentation before text."""
        return line_beginning_re.sub('\t' * level, text) if text else ''

    def create_underline_line(self, text, pad_char):
        """Create a sequence of pad_char characters the same lenght as text."""
        return pad_char * len(text) if text else ''
        
    def underline(self, text, pad_char):
        """Underline text with pad_char characters (-, =, or +).

        Args:
            text (str): the text within a tag, to be underlined
            pad_char (str): the character used for the line (-, =, or +)

        Returns:
            an underlined line of text

        Example:
            >>> import html2md
            >>> html2md.MarkdownConverter().underline("123456789", "=")
            '123456789\\n========='
            >>> html2md.MarkdownConverter().underline("123456789  ", "=")
            '123456789\\n========='
        """
        text = (text or '').rstrip()
        line = self.create_underline_line(text, pad_char)
        return "{}\n{}".format(text, line)


    #-----------------------------------------------------------

    # Conversion functions for specific tags (in alphabetic order): 

    def convert_a(self, el, text):
        """Convert html links.

        Example:
            >>> import html2md
            >>> h = '<a href="a/b/c">abc</a>'
            >>> html2md.markdownify(h)
            '[abc](a/b/c)'
        """
        href = el.get('href')
        title = el.get('title')
        if self.options['autolinks'] and text == href and not title:
            # Shortcut syntax
            return '<%s>' % href
        title_part = ' "%s"' % title.replace('"', r'\"') if title else ''
        return '[%s](%s%s)' % (text or '', href, title_part) if href else text or ''

    def convert_b(self, el, text):
        return self.convert_strong(el, text)

    def convert_blockquote(self, el, text):
        return '\n' + line_beginning_re.sub('> ', text) if text else ''

    def convert_br(self, el, text):
        return '  \n'

    def convert_em(self, el, text):
        return '*%s*' % text if text else ''
    
    def convert_hn(self, n, el, text):
        style = self.options['md_style']
        text = text.rstrip()
        if style == OPENITI:
            return '\n\n### %s %s\n\n' % ("|"*n, text)
        if style == UNDERLINED and n <= 2:
            line = '=' if n == 1 else '-'
            return '\n\n%s\n\n' % self.underline(text, line)
        hashes = '#' * n
        if style == ATX_CLOSED:
            return '\n\n%s %s %s\n\n' % (hashes, text, hashes)
        return '\n\n%s %s\n\n' % (hashes, text)

    def convert_i(self, el, text):
        return self.convert_em(el, text)

    def convert_img(self, el, text):
        """
        Examples:
            >>> import html2md
            >>> h = '<div><img class="figure" src="../Images/figure1.png" /></div>'
            >>> html2md.markdownify(h)
            '![](../Images/figure1.png)'
        """
        
        alt = el.attrs.get('alt', None) or ''
        src = el.attrs.get('src', None) or ''
        title = el.attrs.get('title', None) or ''
        title_part = ' "%s"' % title.replace('"', r'\"') if title else ''
        return '![%s](%s%s)' % (alt, src, title_part)

    def convert_list(self, el, text):
        """Convert ordered and unordered html lists (<ul> and <ol> tags).

        Examples:
            # unordered lists: 
            >>> import html2md
            >>> h = '<ul><li>item1</li><li>item2</li></ul>'
            >>> html2md.markdownify(h)
            '\\n* item1\\n* item2\\n\\n'
            
            # ordered lists:
            >>> import html2md
            >>> h = '<ol><li>item1</li><li>item2</li></ol>'
            >>> html2md.markdownify(h)
            '\\n1. item1\\n2. item2\\n\\n'

            # nested lists:
            ###### TEST FAILS FOR UNKNOWN REASONS
            ##>>> import html2md
            ##>>> h = '<ol><li>item1</li><li>item2:<ul><li>item3</li><li>item4</li></ul></li></ol>'
            ##>>> html2md.markdownify(h)
            ##'\\n1. item1\\n2. item2:\\n\\n\\t* item3\\n\\t* item4\\n\\t\\n\\n'           
        """
        nested = False
        while el:
            if el.name == 'li':
                nested = True
                break
            el = el.parent
        if nested:
            text = '\n' + self.indent(text, 1)
        return '\n' + text + '\n'


    def convert_li(self, el, text):
        """Convert list element tags <li>."""
        parent = el.parent
        if parent is not None and parent.name == 'ol':
            bullet = '%s.' % (parent.index(el) + 1)
        else:
            depth = -1
            while el:
                if el.name == 'ul':
                    depth += 1
                el = el.parent
            bullets = self.options['bullets']
            bullet = bullets[depth % len(bullets)]
        return '%s %s\n' % (bullet, text or '')

    convert_ol = convert_list


    def convert_p(self, el, text):
        """Converts <p> tags.

        Examples:
            >>> import html2md
            >>> h = "<p>abc</p>"
            >>> html2md.markdownify(h)
            '\\n\\n# abc\\n\\n'

            >>> h = "<p>abc</p>"
            >>> html2md.markdownify(h, md_style=ATX)
            '\\n\\nabc\\n\\n'

            >>> h = "<p></p>"
            >>> html2md.markdownify(h, md_style=ATX)
            ''
        """
        if self.options['md_style'] == OPENITI:
            return '\n\n# %s\n\n' % text if text else ''
        else:
            return '\n\n%s\n\n' % text if text else ''


    def convert_table(self, el, text):
        """Wrap tables between double new lines.

        NB: conversion of the tables takes place on the tr level."""
        #print("CONVERTING TABLE")
        return '\n\n' + text + '\n\n'

    def convert_tr(self, el, text):
        """Convert table rows.
        
        NB: rows are processed before the table tag is.
        Spaces to fill out columns are added in post-processing!

        Examples:
            >>> import html2md
            >>> h = '\
            <table>\
              <tr>\
                <th>th1aaa</th><th>th2</th>\
              </tr>\
              <tr>\
                <td>td1</td><td>td2</td>\
              </tr>\
            </table>'
            >>> html2md.markdownify(h)
            '\\n\\n| th1aaa | th2 |\\n| ------ | --- |\\n| td1    | td2 |\\n\\n'
            
            # i.e.,
            # | th1aaa | th2 |
            # | td1    | td2 |
        """
##        def wrap_cell_text(cell_text):
##            while len(cell_text) < self.max_len:
##                cell_text += "ç" # placeholder; will be replaced by spaces later
##            return cell_text
##
##        try:
##            self.max_len
##        except:
##            self.max_len = 0
##        if self.max_len == 0:
##            max_len = 0
##            for desc in el.find_all():
##                if len(desc.text) > max_len:
##                    max_len = len(desc.text)
##            if max_len>30:
##                max_len = 30
##            self.max_len = max_len
                
        t = []
        if el.find('th'): # headers: 
            for th in el.find_all('th'):
                #t.append(wrap_cell_text(th.text))
                t.append(th.text)
            t = "|".join(t)
            return '|{}|\n|{}|\n'.format(t, self.create_underline_line(t, '-'))
        else:
            for td in el.find_all('td'):
                #t.append(wrap_cell_text(td.text))
                t.append(td.text)
            return '|{}|\n'.format("|".join(t))      

    convert_ul = convert_list
        
    def convert_strong(self, el, text):
        """Convert <b> and <strong> tags.

        NB: convert_b refers to this same function

        Examples:
            >>> import html2md
            >>> h = 'abc <strong>def</strong> ghi'
            >>> html2md.markdownify(h)
            'abc **def** ghi'

            >>> import html2md
            >>> h = 'abc <b>def</b> ghi'
            >>> html2md.markdownify(h)
            'abc **def** ghi'
        """
        return '**%s**' % text if text else ''




def markdownify(html, **options):
    return MarkdownConverter(**options).convert(html)


if __name__ == "__main__":
    test_html = """
    <?xml version="1.0" encoding="UTF-8"?>
    <html xml:lang="ar" lang="ar" dir="rtl" xmlns="http://www.w3.org/1999/xhtml" xmlns:epub="http://www.idpf.org/2007/ops">
    <head>
    <meta charset="utf-8" />
    <meta name="viewport" content="initial-scale=2.3, user-scalable=no" />
    <title>TITLE</title>
    <link rel="stylesheet" type="text/css" href="../Style/epub.css" />
    </head>
    <body>
    <div dir="rtl" id="1">
      <h1 dir="rtl" class="title center">H1 TITLE</h1>
      <div dir="rtl" class="subtitle" id="2">SUBTITLE DIV
      </div>
      <div dir="rtl" class="section" id="sect1_34">
        <h4 dir="rtl" class="title center">H4 SECTION TITLE</h4>
        <div class="center" id="4">
          <p>P BLA <span class="quran">QURAN QUOTE</span> P BLABLA</p>
        </div>
      </div>
      <div dir="rtl" class="section" id="sect1_35">
        <h4 dir="rtl" class="title">H4 SECTION TITLE</h4>
        <p>P1
                                    
        </p>
        <p>
                            P2
                        P2b
                                    
        </p>
        <div dir="rtl" class="section" id="sect2_3">
          <h4 dir="rtl" class="title">H4 SUBSECTION2_3 TITLE</h4>
          <p>SUBSECTION2_3 PARAGRAPH</p>
        </div>
      </div>
    </div>
    </body>
    </html>
    """

    import doctest
    doctest.testmod()

    soup = BeautifulSoup(test_html)
    text_div = soup.html.body.div
    #print(text_div)
    #text=markdownify(text_div, md_style=ATX)
    #print("*"*40)
    #print(text)
    
    





"""Convert Epubs from the Hindawi library to OpenITI mARkdown.

Examples:
    >>> from epub_converter_hindawi import HindawiEpubConverter
    >>> from yml2json import yml2json
    >>> folder = "test"
    >>> fn = "26362727.epub"
    >>> hc = HindawiEpubConverter(dest_folder="test/converted")
    >>> meta_fp = "test/hindawi_metadata_man.yml"
    >>> hc.metadata = yml2json(meta_fp, container = {})
    >>> hc.convert_file(os.path.join(folder, fn))
    >>> hc.convert_files_from_folder(folder)

An Epub file is in fact a zipped archive.
The most important element of the archive for conversion purposes
are the folders that contain the html files with the text.
Some Epubs have a table of contents that defines the order
in which these html files should be read.

The HindawiEpubConverter is a subclass of the GenericEpubConverter,
which in turn is a subclass of the GenericConverter
from the generic_converter module:

GenericConverter
    |_ GenericEpubConverter
            |_ HindawiEpubConverter

HindawiConverter's main methods are inherited from the GenericConverter:
* convert_file(source_fp): basic procedure for converting an epub file.
* convert_files_in_folder(source_folder): convert all epub files in the folder
    (calls convert_file)

Methods of both classes:
(methods of GenericConverter are inherited by GenericEpubConverter;
methods of GenericConverter with the same name
in GenericEpubConverter are overwritten by the latter)

| *generic_converter*        | *epub_converter_hindawi* | *epub_converter_hindawi*
|----------------------------|--------------------------|-------------------------
| __init__                   | __init__                 | __init__ 
| convert_files_in_folder    | (inherited)              | (inherited)
| convert file               | (inherited)              | (inherited)
| make_dest_fp               | (inherited - generic!)   | (inherited - generic!)
| get_metadata               | (inherited - generic!)   | get_metadata
| get_data                   | get_data                 | (inherited)
| pre_process                | (inherited)              | (inherited)
| add_page_numbers           | (inherited - generic!)   | (inherited - generic!)
| add_structural_annotations | (inherited - generic!)   | (inherited - generic!) 
| remove_notes               | remove_notes             | (inherited)
| reflow                     | (inherited)              | (inherited)
| add_milestones             | (inherited)              | (inherited)
| post_process               | (inherited - generic!)   | (inherited - generic!) 
| compose                    | (inherited)              | (inherited)
| save_file                  | (inherited)              | (inherited)
|                            | convert_html2md          | convert_html2md
|                            | inspect_epub             | (inherited)
|                            | sort_html_files_by_toc   | (inherited)
|                            | add_unique_tags          | (inherited)


"""

import os

from epub_converter_generic import GenericEpubConverter
import html2md_hindawi
from yml2json import yml2json



class HindawiEpubConverter(GenericEpubConverter):
    def __init__(self, dest_folder=None):
        super().__init__(dest_folder)
        self.toc_fn = "nav.xhtml"

    def convert_html2md(self, html):
        """Use custom html to mARKdown function for Hindawi epubs."""
        text = html2md_hindawi.markdownify(html)
        return text

    def get_metadata(self, source_fp):
        """Custom method to get the metadata of the Hindawi epub file."""
        bookID = os.path.split(source_fp)[1]
        bookID = os.path.splitext(bookID)[0]
        meta_dic = self.metadata[bookID]
        meta = ["#META# {}: {}".format(k,v) for k,v in sorted(meta_dic.items())]
        return self.magic_value + "\n".join(meta) + self.header_splitter


if __name__== "__main__":
    hc = HindawiEpubConverter(dest_folder="test/converted")

    # identify the location of the yml file containing the metadata:
    meta_fp = r"test\hindawi_metadata_man.yml"
    hc.metadata = yml2json(meta_fp, container={})
    
    fp = r"test\26362727.epub"
    hc.convert_file(fp)
    print("converted Hindawi epub", fp)

    hc.convert_files_in_folder("test/hindawi")
    print("converted all epub files in folder", "test/hindawi")




"""Convert a YML file to json.

The yml2json function can be used to convert a yml metadata file to
* a list of dictionaries: useful for
    - conserving the order
    - conserving comments between the records
* a dictionary of dictionaries (key: record id, val: metadata dictionary),
  useful for easy lookup by record id
"""

import json
import re


def prettyprint(obj):
    print(json.dumps(obj, ensure_ascii=False, indent=4, sort_keys=True))


def yml2json(yml_fp, container=[], rec_start="##RECORD",
             id_key="01#BookID###", key_val_splitter=":::: "):
    """Convert a YML file to json.

    Args:
        yml_fp (str): the filepath to the YML file.
        container (obj): an empty list or dict into which the 
            metadata dictionaries will be stored.
            If the container is a list, the order of the records will be stored
            as well as comment lines between the records.
        rec_start (str): the starting characters of a record;
        id_key (str): the yml key that indicates the id of a record
        key_val_splitter (str): the separator used to separate keys and values.

    Returns:
        container (obj): the container filled with the metadata dictionaries.
    """
    spl = key_val_splitter

    with open(yml_fp, mode="r", encoding="utf-8") as file:
        yml_str = file.read()
    blocks = re.split("\n{2,}", yml_str)
    print("The file contains {} blocks".format(len(blocks)))    
    for b in blocks:

        # if the block is a record: build a record dictionary
        
        if b.startswith(rec_start):
            record_lines = b.strip().splitlines()
            r_dic = {line.split(spl)[0]:line.split(spl)[1]
                     for line in record_lines[1:]}
            try:
                container.append(r_dic)
            except:
                container[r_dic[id_key]] = r_dic

        # if the block is a comment in between records,
        # save the comment only if the container is a list:
        
        else:
            try:
                container.append(b)
            except:
                pass
    return container

if __name__ == "__main__":
    meta_fp = r"test/hindawi_metadata_man.yml"
    meta = yml2json(meta_fp, container=[])
    print("A LIST OF DICTIONARIES:")
##    for dic in meta[:2]:
##        try:
##            for k,v in sorted(dic.items()):
##                print(k, v)
##        except:
##            print(dic)
##        print()
    prettyprint(meta[:2])
    print()
    uri_fp = r"test/hindawi_metadata_man.yml"
    uri_data = yml2json(uri_fp, container={})
    print("A DICTIONARY OF DICTIONARIES:")
    for key, dic in sorted(list(uri_data.items()))[:2]:
        for k,v in sorted(dic.items()):
            print(k, v)
        print()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="openiti",
    version="0.0.2",
    author="Soheil Merchant, Maxim Romanov, Masoumeh Seydi, Peter Verkinderen",
    author_email="peter.verkinderen@gmail.com",
    description="A package for dealing with the openITI corpus",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://https://github.com/OpenITI",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.4',
)



import sys
import requests
import re
import subprocess


def get_total_repos(group, name, path_pattern):
    repo_urls = []
    page = 1
    while True:
        url = 'https://api.github.com/{0}/{1}/repos?per_page=100&page={2}'
        r = requests.get(url.format(group, name, page))
        print(url.format(group, name, page))
        if r.status_code == 200:
            rdata = r.json()
            for repo in rdata:
                if re.compile(path_pattern).search(repo['full_name']):
                    repo_urls.append(repo['clone_url'])
            if (len(rdata) >= 100):
                page = page + 1
            else:
                print('Found {0} repos.'.format(len(repo_urls)))
                break
        else:
            print(r)
            return False
    return repo_urls


def clone_repos(all_repos, clone_dir):
    print('Cloning...')
    counter = 1
    total_repo_nr = str(len(all_repos))
    for repo in all_repos:
        path_to_clone = clone_dir #+ repo.split("/")[-1]
        print ('\nRepository ' + str(counter) + ' of ' + total_repo_nr + '\n')
        print(path_to_clone)
        p = subprocess.Popen(["git", "clone", repo], cwd=path_to_clone)
        p.wait()
        # os.system('git clone ' + repo + ' ' + path_to_clone + repo.split("/")[-1])
        counter += 1


if __name__ == '__main__':

    # path_pattern = re.compile("\d{4}AH$")
    type = input("Organization or User (orgs/users)?: ")
    org_name = input("Enter the organization name: ")
    path_pattern = input("Enter the regex pattern in repo URL: ")
    clone_dir = input("Enter the path to clone: ")
    # print("arg: " ,sys.argv)
    if len(sys.argv) > 0:
        total = get_total_repos(type, org_name, path_pattern)
        # for clone_OpenITI.py orgs OpenITI usage. Then, len(sys.argv) have to be >2!
        # total = get_total_repos(sys.argv[1], sys.argv[2])
        if total:
            clone_repos(total, clone_dir)
        else:
            print("No repository found!")
    else:
        print('In Terminal, cd to the destination directory in which you want to clone the repos.\n '
              'then run the script:\n'
              '\tpython3 path/to/script/clone_OpenITI.py OpenITI\n'
              'OpenITI is the organization name!')


import re
import os
import sys

# TODO: update the paths using import
sys.path.append("Z:/Documents/GitProjects/PythonFunctions")
sys.path.append("/Users/romanov/Documents/c.GitProjects/PythonFunctions")

sum_r = """
    Min.  1st Qu.   Median     Mean  3rd Qu.     Max. 
     150    32700   105300   306700   266700 10410000 
"""


# loop through all subfolders
# check all files ending with [-ara1]
# count the overall number of files
# count word freqs for all of them (build a table)
# count the overall length in words
# count word frequencies of only longest texts in each text group


def get_len(file_name):
    with open(file_name, "r", encoding="utf8") as f1:
        f1 = f1.read()
        fr = len(re.findall("\w+", f1))
        return fr


def generate_stats(in_dir):
    stats = []
    for root, dirs, filenames in os.walk(in_dir):
        for f in filenames:
            if f.endswith("-ara1"):
                path_to_file = os.path.join(root, f)
                # print(path_to_file)

                fr1 = get_len(path_to_file)
                fr2 = "{:,}".format(fr1)
                stats.append("%010d\t%s\t%s" % (fr1, f, fr2))

                print("%010d\t%s\t%s" % (fr1, f, fr2))

    with open("raw_len_stats.txt", "w", encoding="utf8") as f9:
        f9.write("\n".join(stats))


# generateStats()
def ref(sum_re):
    sum_re = re.sub(" Q", "_Q", sum_re)
    sum_re = re.sub(" +", "\t", sum_re)
    sum_re = re.sub("_", " ", sum_re)
    sum_re = re.sub("\n\n", "", sum_re)
    sum_re = re.sub(r"(^|\n)\t", r"\1", sum_re)

    head = sum_re.split("\n")[1].split("\t")
    vals = sum_re.split("\n")[2].split("\t")

    # TABLE
    doc = "\n\n## Summary statistics on the lengths of texts in the corpus\n\n"

    doc += "| | Words  | Pages (300 w/p) |\n"
    doc += "|:--- | ------:| -----:|\n"

    for i in range(0, 6):
        doc += "| *%s* | %s | %s |\n" % (head[i], "{:,}".format(int(vals[i])), "{:,}".format(int(vals[i]) // 300))
    return doc


def process_stats(len_stats_file, collect_stats_file):
    doc = "## Statistics on the corpus\n\n"
    doc += "| **Category** | **Stats** |\n"
    doc += "|:--- | ------:|\n"

    with open(len_stats_file, "r", encoding="utf8") as f1:
        stats = f1.read().split("\n")

        uri_dic = {}
        uri_lis = []
        for i in stats:
            i = i.split("\t")
            uri_dic[i[1]] = int(i[0])
            uri_lis.append(i[1])

        # count authors and books
        authors = []
        books = []
        for a in uri_lis:
            a = a.split(".")
            authors.append(a[0])
            books.append(a[0] + "." + a[1])

        # count words all and by unique titles
        uni_dic = {}
        words_all = 0
        for k, v in uri_dic.items():
            words_all += v
            new_k = ".".join(k.split(".")[:2])
            if new_k in uni_dic:
                if uni_dic[new_k] < v:
                    uni_dic[new_k] = v
            else:
                uni_dic[new_k] = v

        words_max = 0
        lengths = []
        list_uri_freq = []
        list_freq_uri = []
        for k, v in uni_dic.items():
            words_max += v
            lengths.append(str(v))
            list_uri_freq.append("%s\t%09d" % (k, v))
            list_freq_uri.append("%09d\t%s" % (v, k))

        print("Run in R to get summary stats:\n\nsummary(c(%s))" % ",".join(lengths))

        print("\n\n")

        doc += "| Number of text files | %d |\n" % len(stats)
        doc += "| Number of books      | %d |\n" % len(list(set(books)))
        doc += "| Number of authors    | %d |\n" % len(list(set(authors)))
        doc += "| Length in words (all)| %s |\n" % "{:,}".format(words_all)
        doc += "| Length in pages (300 w/p)| %s |\n" % "{:,}".format(words_all // 300)
        doc += "| Length in words (unique) | %s |\n" % "{:,}".format(words_max)
        doc += "| Length in pages (300 w/p)| %s |\n" % "{:,}".format(words_max // 300)

        doc += ref(sum_r)

        # table of texts by lengths

        doc += "\n\n## Texts by length (duplicates excluded)\n\n"

        doc += "| Num | TextGroup URI | Words  | Pages (300 w/p) |\n"
        doc += "| --: | :--- | ------:| -----:|\n"

        counter = 0

        for i in sorted(list_freq_uri, reverse=True):
            counter += 1
            i = i.split("\t")
            doc += "| %d | %s | %s | %s |\n" % (
                counter, i[1], "{:,}".format(int(i[0])), "{:,}".format(int(i[0]) // 300))

        # table of texts by chronology

        doc += "\n\n## Texts in chronological order (duplicates excluded)\n\n"
        doc += "| TextGroup URI | Words  | Pages (300 w/p) |\n"
        doc += "|:--- | ------:| -----:|\n"

        for i in sorted(list_uri_freq, reverse=False):
            i = i.split("\t")
            doc += "| %s | %s | %s |\n" % (i[0], "{:,}".format(int(i[1])), "{:,}".format(int(i[1]) // 300))

        with open(collect_stats_file, "w", encoding="utf8") as f9:
            f9.write(doc)



"""Functions to read and write yaml files.

OpenITI metadata is stored in yaml files. 
"""

import re
import textwrap

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)

from openiti.helper.templates import author_yml_template, book_yml_template, \
                                     version_yml_template



def ymlToDic(yml_str):
    """Convert a yml string into a dictionary"""
    data = re.sub("\n+$", "", yml_str)
    data = re.sub("-\n[ \t]+", "-", data)
    data = re.sub("\n[ \t]+", " ", data)
    data = data.split("\n")
    dic = dict()
    for d in data:
        d = re.split(r"(^[\w#]+:)", d)
        dic[d[1]] = d[2].strip()
    return dic

def readYML(fp):
    """Read a yml file and convert it into a dictionary"""
    with open(fp, "r", encoding="utf8") as file:
        return ymlToDic(file.read())

def dicToYML(dic):
    """Convert a dictionary into a yml string"""
    data = []
    for k,v in dic.items():
        if k.endswith(":"):
            i = k + " " + str(v)
        else:
            i = k + ": " + str(v)
        if "#URI#" in i:
            pass
        else:
            i = "\n    ".join(textwrap.wrap(i, 72))
        data.append(i)

    data = "\n".join(sorted(data))
    return(data)
